open Utils
open Binding
open Resumable

type 
  lsort =  LType
	  |LSet
	  |LProp
and
  lterm =  LSort of lsort
	  |LVar of lvar
	  |LLambda of string option * lsort inferred * lterm * lterm
	  |LPi of string option * lsort inferred * lterm * lterm
	  |LApp of lterm * lterm
	  |LInd of linddef
	  |LCtor of lterm * int
	  |LElim of lterm (* result *) * lterm (* ind. term *) * lterm list (* branches *)

	  (* added for meta variables *)
	  |LModal of lmodalterm * lsubst

	  (* added for term lists (contexts and substitutions) *)
	  |LTermList of lctxdesc

	  |LInfer of lunifvar ref * (lterm -> lterm)
and
  lvar =    LBVar of int  (* bound variables: deBruijn indexes *)
          | LFVar of int  (* free variables: deBruijn levels *)
	  | LNVar of string (* named variables: names in definition environments (maps) *)
and
  lmodalterm =   LTermInCtx of lctx * lterm
	       | LFMeta of int
	       | LBMeta of int 
	       | LNMeta of string
and
  lctxdesc =     LFCtx of int
	       | LBCtx of int
	       | LCtxAsList of lctx
and
  linddef = LIndDef of string option * larity * lconstr list (* inddef := Ind(X : Arity) { Constr_0 | ... | Constr_n } *)
and
  larity = LArity of (string option * lterm) list * lsort      (* arity  := Pi x_0 : M_0. ... Pi x_n : M_n. s *)
and
  lconstr = LConstr of (string option * lcparam) list * lterm list  (* constr := Pi x_0 : Cparam_0. ... Pi x_n : Cparam_n.X m_0 ... m_n *)
and
  lcparam =   LParNonrec of lterm  (* nonrecursive parameter for constructor of X: Pi x : M. ... with X not in fv(M) *)
            | LParStrictPos of (string option * lterm) list * lterm list (* recursive parameter with strictly positive occurence of X:
							  (P -> ...) with P = Pi x_0 : M_0. ... Pi x_n : M_n. X m_0 ... m_n with X not in fv(M_i |_| m_i) *)
and
  lsubst = lterm list
and
  lctx = (string option * lterm) list
and
  lunifvar = lterm unifvar

let _lunifcount    = ref 0
let _lifinfer      = (function LInfer(l,_) -> (fun y n -> y l) | _ -> (fun y n -> n))
let _linject  l f  = LInfer(l,f)
let mk_lunifvar () = mk_unifvar _lunifcount
let match_lunif    = match_unif _lifinfer
let lunifvar_type_unify unification = unifvar_type_unify unification _lifinfer _linject _lunifcount


type lterm_env = lterm list (* free variables *) * lterm_defenv (* def. environment *) * lmodalterm list (* meta variables *) * unit list (* ctx variables *)
and  lterm_defenv = (lterm (* type *)) dict * (lmodalterm (* type *) * lmodalterm option (*term *)) dict


(* Map *)

let lterm_map ?(lsort=fun s -> LSort(s))
              ?(lbvar=fun i -> LVar(LBVar i))
	      ?(lfvar=fun i -> LVar(LFVar i))
	      ?(lnvar=fun s -> LVar(LNVar s))
	      ?(llambda=fun s k t1 t2 rt1 rt2 -> LLambda(s, k, rt1, rt2))
	      ?(lpi=fun s k t1 t2 rt1 rt2 -> LPi(s, k, rt1, rt2))
	      ?(lapp=fun t1 t2 rt1 rt2 -> LApp(rt1, rt2))
	      ?(lctor=fun t1 i1 rt1 -> LCtor(rt1, i1))
	      ?(lelim=fun t1 t2 t3s rt1 rt2 rt3s -> LElim(rt1,rt2,rt3s))
	      ?(lsubstelem=fun t1 rt1 -> [rt1])
	      ?(lfctx=fun i1 -> LFCtx(i1))
	      ?(lbctx=fun i1 -> LBCtx(i1))
	      ?(lfmeta=fun i1 -> LFMeta(i1))
	      ?(lbmeta=fun i1 -> LBMeta(i1))
	      ?(lnmeta=fun s rf -> LNMeta(s))
	      ?(lctxelem=fun s t1 rt1 -> [(s,rt1)])
	      ?(lmodal=fun mt1 t2s rmt1 rt2s -> LModal(rmt1, rt2s))
	      ?(lterminctx=fun ctx lt rctx rlt -> LTermInCtx(rctx, rlt))
	      ?(lctxaslist=fun ctx rctx -> LCtxAsList(rctx))
	      ?(linfer=fun l f rf -> LInfer(l,rf))
    =

  let mapsnd f = List.map (fun (a,b) -> (a, f b)) in
  let rec ltrec = function
      LSort(s) -> lsort s
    | LVar(LBVar i) -> lbvar i
    | LVar(LFVar i) -> lfvar i
    | LVar(LNVar i) -> lnvar i
    | LLambda(s1, k, t1, t2) -> llambda s1 k t1 t2 (ltrec t1) (ltrec t2)
    | LPi(s1, k, t1, t2) -> lpi s1 k t1 t2 (ltrec t1) (ltrec t2)
    | LApp(t1, t2) -> lapp t1 t2 (ltrec t1) (ltrec t2)
    | LInd(i1) -> LInd(lind i1)
    | LCtor(t1,i1) -> lctor t1 i1 (ltrec t1)
    | LElim(t1,t2,t3s) -> lelim t1 t2 t3s (ltrec t1) (ltrec t2) (List.map ltrec t3s)
    | LModal(mt1,t2s) -> lmodal mt1 t2s (modrec mt1) (List.concat (List.map substelemrec t2s))
    | LTermList(ct1) -> LTermList(ltermlist ct1)
    | LInfer(l, f) -> linfer l f (compose ltrec f)
  and substelemrec t1 = lsubstelem t1 (ltrec t1)
  and ctxelemrec (s, t1) = lctxelem s t1 (ltrec t1)
  and modrec = function
      LFMeta(i) -> lfmeta i
    | LBMeta(i) -> lbmeta i
    | LNMeta(s) -> lnmeta s modrec
    | LTermInCtx(ct1,t1) -> lterminctx ct1 t1 (List.concat (List.map ctxelemrec ct1)) (ltrec t1)
  and ltermlist = function
      LFCtx(i) -> lfctx i
    | LBCtx(i) -> lbctx i
    | LCtxAsList(l) -> lctxaslist l (List.concat (List.map ctxelemrec l))
  and lind = function
      LIndDef(name, LArity(l1, s2), constrlist) ->
	LIndDef(name, LArity(mapsnd ltrec l1, s2), List.map lconstr constrlist)
  and lconstr = function
      LConstr(l1, t2s) -> LConstr(mapsnd lcparam l1, List.map ltrec t2s)
  and lcparam = function
      LParNonrec(l1) -> LParNonrec(ltrec l1)
    | LParStrictPos(l1,l2) -> LParStrictPos(mapsnd ltrec l1, List.map ltrec l2)
  in
    ltrec

let remove_infer t =
  lterm_map ~linfer:(fun l f rf -> (match !(match_lunif l) with
					Inst(t,_) -> rf t
				      | _ -> failwith "couldn't infer this variable!")) t

let remove_infer_lmodal mt = 
  match remove_infer (LModal(mt,[])) with
      LModal(mt,_) -> mt
    | _ -> failwith "remove_infer_lmodal"

let remove_infer_lctxdesc ctx = 
  match remove_infer (LTermList(ctx)) with
    LTermList(ctx) -> ctx
    | _ -> failwith "remove_infer_lctxdesc"


(* Binding *)

(* this can be used to apply a varmap to a list of
   terms, each of which is a binder *)
		
let list_varmap f_varmap fextend start list =
  let rec list_varmap_aux i start (list : 'b list) =
    match list with
	[] -> []
      | hd :: tl -> (f_varmap (fextend start i) hd) :: list_varmap_aux (i+1) start tl
  in
    list_varmap_aux 0 start list

(* this is to be used when the number of binders cannot be
   known beforehand *)
let list_varmap' f_varmap fextend start list =
  let rec list_varmap_aux start (list : 'b list) =
    match list with
	[] -> [], start
      | hd :: tl -> 
	  let hd', s' = f_varmap start hd in
	  let tl', s'' = list_varmap_aux s' tl in
	    (hd' :: tl'), s''
  in
    list_varmap_aux start list


let varmaps fbound ffree =
  let fextend s i = s + i in
  let rec lterm_varmap start (t : lterm) =
    match t with
	LVar (LBVar i) -> fbound start i
      | LVar (LFVar i) -> ffree start i
      | LLambda (s, k, lt1, lt2)  -> LLambda(s, k, lterm_varmap start lt1, lterm_varmap (fextend start 1) lt2)
      | LPi (s, k, lt1, lt2)      -> LPi(s, k, lterm_varmap start lt1, lterm_varmap (fextend start 1) lt2)
      | LApp (lt1, lt2)        -> LApp(lterm_varmap start lt1, lterm_varmap start lt2)
      | LInd (idef)            -> LInd(inddef_varmap start idef)
      | LCtor (lt1, i)         -> LCtor(lterm_varmap start lt1, i)
      | LElim(lt1,lt2,ltl)     -> LElim(lterm_varmap start lt1, lterm_varmap start lt2, List.map (lterm_varmap start) ltl)
      | LModal(lt, subst)      -> LModal(modal_varmap start lt, List.map (lterm_varmap start) subst)
      | LTermList(tmlist)      -> LTermList(tmlist_varmap start tmlist)
      | LInfer(l,f)            -> 
	  (match !(match_lunif l) with
	       Inst(t,_) -> lterm_varmap start (f t)
	     | _ -> LInfer(l, compose (lterm_varmap start) f))
      | t -> t
  and inddef_varmap start (t : linddef) =
    match t with
	LIndDef(s, arity, constrlist) ->
	  LIndDef(s, arity_varmap start arity, List.map (constr_varmap start) constrlist)
  and arity_varmap start (t : larity) =
    match t with
	LArity(termlist, sort) ->
	  LArity(stringlist (list_varmap lterm_varmap fextend start) termlist, sort)
  and constr_varmap start (t : lconstr) =
    match t with
	LConstr(paramlist, arglist) ->
	  (let nparam = List.length paramlist in
	     LConstr(stringlist (list_varmap cparam_varmap fextend start) paramlist,
		     List.map (lterm_varmap (fextend start nparam)) arglist))
  and cparam_varmap start (t : lcparam) =
    match t with
	LParNonrec(lt) -> LParNonrec(lterm_varmap start lt)
      | LParStrictPos(paramlist, arglist) ->
	  (let nparam = List.length paramlist in
	     LParStrictPos(stringlist (list_varmap lterm_varmap fextend start) paramlist,
			   List.map (lterm_varmap (fextend start nparam)) arglist))
  and modal_varmap start (t : lmodalterm) =
    match t with
	LTermInCtx(ctx, lt) ->
	  (let n = List.length ctx in
	     LTermInCtx(stringlist (list_varmap lterm_varmap fextend start) ctx,
			lterm_varmap (fextend start n) lt))
      | LFMeta(i) -> LFMeta(i)
      | LBMeta(i) -> LBMeta(i)
      | LNMeta(s) -> LNMeta(s)
  and tmlist_varmap start (t : lctxdesc) =
    match t with
	LFCtx(i) -> t
      | LBCtx(i) -> t
      | LCtxAsList(l) -> LCtxAsList(stringlist (list_varmap lterm_varmap fextend start) l)
  and ctx_varmap start (t : lctx) =
    stringlist (list_varmap lterm_varmap fextend start) t
  in
    (lterm_varmap, inddef_varmap, arity_varmap, constr_varmap, cparam_varmap, modal_varmap, tmlist_varmap, ctx_varmap)
                                            
let lterm_varmap fbound ffree   = let (a,_,_,_,_,_,_,_) = varmaps fbound ffree in a
let inddef_varmap fbound ffree  = let (_,a,_,_,_,_,_,_) = varmaps fbound ffree in a
let arity_varmap fbound ffree   = let (_,_,a,_,_,_,_,_) = varmaps fbound ffree in a
let constr_varmap fbound ffree  = let (_,_,_,a,_,_,_,_) = varmaps fbound ffree in a
let cparam_varmap fbound ffree  = let (_,_,_,_,a,_,_,_) = varmaps fbound ffree in a
let modal_varmap fbound ffree   = let (_,_,_,_,_,a,_,_) = varmaps fbound ffree in a
let tmlist_varmap fbound ffree  = let (_,_,_,_,_,_,a,_) = varmaps fbound ffree in a
let ctx_varmap fbound ffree     = let (_,_,_,_,_,_,_,a) = varmaps fbound ffree in a

module BindLterm = Binding(struct
			     type tvar = lterm
			     type t = lterm
			     type tenv = int
			     let  bound i = i
			     let  varmap f1 f2 = lterm_varmap f1 f2 0
			     let  bvar   i = LVar (LBVar i)
			     let  fvar   i = LVar (LFVar i)
			   end)
module BindLtermS = (struct
		       let subst_bound_list = BindLterm.subst_bound_list_given_shift BindLterm.shift_bound
		       let subst_bound      = BindLterm.subst_bound_given_list subst_bound_list
		     end)

module BindInddef = Binding(struct
			     type t = linddef
			     type tvar = lterm
			     type tenv = int
			     let  bound i = i
			     let  varmap f1 f2 = inddef_varmap f1 f2 0
			     let  bvar   i = LVar (LBVar i)
			     let  fvar   i = LVar (LFVar i)
			   end)
module BindInddefS = (struct
			let subst_bound_list = BindInddef.subst_bound_list_given_shift BindLterm.shift_bound
			let subst_bound      = BindInddef.subst_bound_given_list subst_bound_list
		      end)

module BindArity = Binding(struct
			     type t = larity
			     type tvar = lterm
			     type tenv = int
			     let  bound i = i
			     let  varmap f1 f2 = arity_varmap f1 f2 0
			     let  bvar   i = LVar (LBVar i)
			     let  fvar   i = LVar (LFVar i)
			   end)

module BindConstr = Binding(struct
			     type t = lconstr
			     type tvar = lterm
			     type tenv = int
			     let  bound i = i
			     let  varmap f1 f2 = constr_varmap f1 f2 0
			     let  bvar   i = LVar (LBVar i)
			     let  fvar   i = LVar (LFVar i)
			   end)

module BindCparam = Binding(struct
			     type t = lcparam
			     type tvar = lterm
			     type tenv = int
			     let  bound i = i
			     let  varmap f1 f2 = cparam_varmap f1 f2 0
			     let  bvar   i = LVar (LBVar i)
			     let  fvar   i = LVar (LFVar i)
			   end)

module BindModal = Binding(struct
			     type t = lmodalterm
			     type tvar = lterm
			     type tenv = int
			     let  bound i = i
			     let varmap f1 f2 = modal_varmap f1 f2 0
			     let bvar i = LVar (LBVar i)
			     let fvar i = LVar (LFVar i)
			   end)

module BindCtx = Binding(struct
			   type t = lctx
			   type tvar = lterm
			   type tenv = int
			   let  bound i = i
			   let varmap f1 f2 = ctx_varmap f1 f2 0
			   let bvar i = LVar (LBVar i)
			   let fvar i = LVar (LFVar i)
			 end)
module BindCtxS = (struct
		     let subst_bound_list = BindCtx.subst_bound_list_given_shift BindLterm.shift_bound
		     let subst_bound      = BindCtx.subst_bound_given_list subst_bound_list
		   end)

module BindTermlist = Binding(struct
				type t = lctxdesc
				type tvar = lterm
				type tenv = int
				let  bound i = i
				let varmap f1 f2 = tmlist_varmap f1 f2 0
				let bvar i = LVar (LBVar i)
				let fvar i = LVar (LFVar i)
			      end)

exception ModalVarmap of lmodalterm * lterm list
exception ModalVarmapHandle of lterm

(* I have started implementing a hack using resumable exceptions in order to
   flatten modal substitutions as soon as they happen *)

let metamaps fbound ffree =
  let rec lterm_varmap start (t : lterm) =
    match t with
      | LLambda (s, k, lt1, lt2)  -> LLambda(s, k, lterm_varmap start lt1, lterm_varmap start lt2)
      | LPi (s, k, lt1, lt2)      -> LPi(s, k, lterm_varmap start lt1, lterm_varmap start lt2)
      | LApp (lt1, lt2)        -> LApp(lterm_varmap start lt1, lterm_varmap start lt2)
      | LInd (idef)            -> LInd(inddef_varmap start idef)
      | LCtor (lt1, i)         -> LCtor(lterm_varmap start lt1, i)
      | LElim(lt1,lt2,ltl)     -> LElim(lterm_varmap start lt1, lterm_varmap start lt2, List.map (lterm_varmap start) ltl)
      | LModal(lt, subst)      -> 
	 (let lt' = modal_varmap start lt in
	  let subst' = List.map (lterm_varmap start) subst in
	  (* rraise (ModalVarmap (lt', subst')) *)
	  (* (fun e -> try raise e with ModalVarmapHandle(lt'') -> lt'') *)
	  LModal(lt',subst')
	 )
      | LTermList(ctxdesc)     -> LTermList(ctxdesc_varmap start ctxdesc)
      | LInfer(l,f) ->
	  (match !(match_lunif l) with
	       Inst(t,_) -> lterm_varmap start (f t)
	     | _ -> LInfer(l, compose (lterm_varmap start) f))
      | t -> t
  and inddef_varmap start (t : linddef) =
    match t with
	LIndDef(s, arity, constrlist) ->
	  LIndDef(s, arity_varmap start arity, List.map (constr_varmap start) constrlist)
  and arity_varmap start (t : larity) =
    match t with
	LArity(termlist, sort) ->
	  LArity(stringlist (List.map (lterm_varmap start)) termlist, sort)
  and constr_varmap start (t : lconstr) =
    match t with
	LConstr(paramlist, arglist) ->
	  (LConstr(stringlist (List.map (cparam_varmap start)) paramlist,
		     List.map (lterm_varmap start) arglist))
  and cparam_varmap start (t : lcparam) =
    match t with
	LParNonrec(lt) -> LParNonrec(lterm_varmap start lt)
      | LParStrictPos(paramlist, arglist) ->
	  (LParStrictPos(stringlist (List.map (lterm_varmap start)) paramlist,
			   List.map (lterm_varmap start) arglist))
  and modal_varmap start (t : lmodalterm) =
    match t with
	LTermInCtx(ctx, lt) ->
	  (LTermInCtx(stringlist (List.map (lterm_varmap start)) ctx,
		      lterm_varmap start lt))
      | LFMeta(i) -> ffree start i
      | LBMeta(i) -> fbound start i
      | LNMeta(s) -> LNMeta(s)
  and ctxdesc_varmap start (t : lctxdesc) =
    match t with
	LCtxAsList(l) -> LCtxAsList(stringlist (List.map (lterm_varmap start)) l)
      | t -> t
  in
    (lterm_varmap, inddef_varmap, arity_varmap, constr_varmap, cparam_varmap, modal_varmap, ctxdesc_varmap)

let lterm_metamap fbound ffree   = let (a,_,_,_,_,_,_) = metamaps fbound ffree in a
let inddef_metamap fbound ffree  = let (_,a,_,_,_,_,_) = metamaps fbound ffree in a
let arity_metamap fbound ffree   = let (_,_,a,_,_,_,_) = metamaps fbound ffree in a
let constr_metamap fbound ffree  = let (_,_,_,a,_,_,_) = metamaps fbound ffree in a
let cparam_metamap fbound ffree  = let (_,_,_,_,a,_,_) = metamaps fbound ffree in a
let modal_metamap fbound ffree   = let (_,_,_,_,_,a,_) = metamaps fbound ffree in a
let ctxdesc_metamap fbound ffree = let (_,_,_,_,_,_,a) = metamaps fbound ffree in a


exception CtxVarmap1 of lctx
exception CtxVarmap2 of lctx * lterm
exception CtxVarmap1Handle of lctxdesc
exception CtxVarmap2Handle of lctx * lterm

let ctxmaps fbound ffree =
  let rec lterm_varmap start (t : lterm) =
    match t with
      | LLambda (s, k, lt1, lt2)  -> LLambda(s, k, lterm_varmap start lt1, lterm_varmap start lt2)
      | LPi (s, k, lt1, lt2)      -> LPi(s, k, lterm_varmap start lt1, lterm_varmap start lt2)
      | LApp (lt1, lt2)        -> LApp(lterm_varmap start lt1, lterm_varmap start lt2)
      | LInd (idef)            -> LInd(inddef_varmap start idef)
      | LCtor (lt1, i)         -> LCtor(lterm_varmap start lt1, i)
      | LElim(lt1,lt2,ltl)     -> LElim(lterm_varmap start lt1, lterm_varmap start lt2, List.map (lterm_varmap start) ltl)
      | LModal(lt, subst)      -> LModal(modal_varmap start lt, List.map (lterm_varmap start) subst)
      | LTermList(tmlist)      -> LTermList(tmlist_varmap start tmlist)
      | LInfer(l,f) ->
	  (match !(match_lunif l) with
	       Inst(t,_) -> lterm_varmap start (f t)
	     | _ -> LInfer(l, compose (lterm_varmap start) f))
      | t -> t
  and inddef_varmap start (t : linddef) =
    match t with
	LIndDef(s, arity, constrlist) ->
	  LIndDef(s, arity_varmap start arity, List.map (constr_varmap start) constrlist)
  and arity_varmap start (t : larity) =
    match t with
	LArity(termlist, sort) ->
	  LArity(stringlist (List.map (lterm_varmap start)) termlist, sort)
  and constr_varmap start (t : lconstr) =
    match t with
	LConstr(paramlist, arglist) ->
	  (LConstr(stringlist (List.map (cparam_varmap start)) paramlist,
		     List.map (lterm_varmap start) arglist))
  and cparam_varmap start (t : lcparam) =
    match t with
	LParNonrec(lt) -> LParNonrec(lterm_varmap start lt)
      | LParStrictPos(paramlist, arglist) ->
	  (LParStrictPos(stringlist (List.map (lterm_varmap start)) paramlist,
			   List.map (lterm_varmap start) arglist))
  and modal_varmap start (t : lmodalterm) =
    match t with
	LTermInCtx(ctx, lt) ->
	  let ctx' = stringlist (List.map (lterm_varmap start)) ctx in
	  let lt' = lterm_varmap start lt in
	  (* rraise (CtxVarmap2 (ctx',lt')) *)
	  (* (fun e -> try raise e with CtxVarmap2Handle(ctx'',lt'') -> LTermInCtx(ctx'',lt'')) *)
	  LTermInCtx(ctx', lt')
      | t -> t
  and tmlist_varmap start (t : lctxdesc) =
    match t with
	LFCtx(i) -> ffree start i
      | LBCtx(i) -> fbound start i
      | LCtxAsList(lt) -> 
	let lctx' = stringlist (List.map (lterm_varmap start)) lt in
	(* rraise (CtxVarmap1 (lctx')) *)
	(* (fun e -> try raise e with CtxVarmap1Handle(lctxdesc') -> lctxdesc') *)
	LCtxAsList(lctx')
  in
    (lterm_varmap, inddef_varmap, arity_varmap, constr_varmap, cparam_varmap, modal_varmap, tmlist_varmap)

let lterm_ctxmap fbound ffree  = let (a,_,_,_,_,_,_) = ctxmaps fbound ffree in a
let modal_ctxmap fbound ffree  = let (_,_,_,_,_,a,_) = ctxmaps fbound ffree in a
let tmlist_ctxmap fbound ffree = let (_,_,_,_,_,_,a) = ctxmaps fbound ffree in a


(* flattenings *)
let modal_apply_subst modal subst = 
    match modal with
	LTermInCtx(ctx,lt) -> 
	  let ctxN = List.length ctx in
	  let substN = List.length subst in
	  let lt' = (if ctxN < substN then
		       BindLterm.shift_bound (substN - ctxN) lt
		     else
		       lt) in
	  BindLtermS.subst_bound_list subst lt'
      | meta -> LModal(meta, subst)

(* --- this is already in logic_core *)
(*
let replace_fN_in_substs n m t =
  lterm_map ~lfvar:(fun i -> LVar (LFVar (if i > n then i + (m - 1) else i)))
            ~lsubstelem:(fun tm res ->
			   match tm with
			       LVar(LFVar n) -> List.map (fun i -> LVar(LFVar (i+n))) (increasing m)
			     | _ -> [res])
    t

let rec flatten_term_in_context ctx t =
  try
    nowarn let (before, (_, LTermList(LCtxAsList(l))), after), index = ExtList.find_partition_index (function (_, LTermList(LCtxAsList(_))) -> true | _ -> false) ctx in
    let m = List.length l in
    nowarn let LTermInCtx(after', t') = BindModal.open_up ~howmany:(index+1) 0 (LTermInCtx(after, t)) in
    let after', t' = List.map (fun (s,t) -> (s,replace_fN_in_substs index m t)) after', replace_fN_in_substs index m t' in
    nowarn let LTermInCtx(after', t') = BindModal.close_down ~howmany:(index+m) (index+m) (LTermInCtx(after',t')) in
      flatten_term_in_context (List.append before (List.append l after')) t'
  with ExtList.NotFoundPartition -> (ctx, t)

let flatten_context ctx =
  let ctx', _ = flatten_term_in_context ctx (LSort(LSet)) in ctx'
*)

(* handling the exceptions now *)

(*
let default_varmap_handler f =
  rhandle (function v ->
    try raise v
    with 
	ModalVarmap(lt,subst) -> resume (ModalVarmapHandle(LModal(lt,subst)))
      | CtxVarmap1(ctx) -> resume (CtxVarmap1Handle(LCtxAsList(ctx)))
      | CtxVarmap2(ctx,lt) -> resume (CtxVarmap2Handle(ctx,lt)))
    f

let default_varmap_handler f =
  rhandle (function v ->
    try raise v
    with 
	ModalVarmap(lt,subst) ->
	  resume (ModalVarmapHandle(modal_apply_subst lt subst))
      | CtxVarmap1(ctx) ->
	  resume (CtxVarmap1Handle(LCtxAsList(flatten_context ctx)))
      | CtxVarmap2(ctx,lt) -> 
	let ctx', lt' = flatten_term_in_context ctx lt in
	resume(CtxVarmap2Handle(ctx',lt')))
    f

(* we could still enable the full maps here, so that they are used only
   when needed *)
let modal_metamap a b c d = default_varmap_handler (fun () -> modal_metamap a b c d)
let lterm_metamap a b c d = default_varmap_handler (fun () -> lterm_metamap a b c d)
let ctxdesc_metamap a b c d = default_varmap_handler (fun () -> ctxdesc_metamap a b c d)

let lterm_ctxmap a b c d = default_varmap_handler (fun () -> lterm_ctxmap a b c d)
let modal_ctxmap a b c d = default_varmap_handler (fun () -> modal_ctxmap a b c d)
let tmlist_ctxmap a b c d = default_varmap_handler (fun () -> tmlist_ctxmap a b c d)
*)

let default_varmap_handler f = f ()

module MetabindLterm = Binding(struct
				 type tvar = lmodalterm
				 type t = lterm
				 type tenv = int
				 let  bound  i = i
				 let  varmap f1 f2 = lterm_metamap f1 f2 0
				 let  bvar   i = LBMeta i
				 let  fvar   i = LFMeta i
			       end)

module MetabindModal = Binding(struct
			     type t = lmodalterm
			     type tvar = lmodalterm
			     type tenv = int
			     let  bound i = i
			     let varmap f1 f2 = modal_metamap f1 f2 0
			     let bvar i = LBMeta i
			     let fvar i = LFMeta i
			   end)

module MetabindCtxdesc = Binding(struct
			     type t = lctxdesc
			     type tvar = lmodalterm
			     type tenv = int
			     let  bound i = i
			     let varmap f1 f2 = ctxdesc_metamap f1 f2 0
			     let bvar i = LBMeta i
			     let fvar i = LFMeta i
			   end)


module CtxbindLterm = Binding(struct
				type tvar = lctxdesc
				type t = lterm
				type tenv = int
				let  bound  i = i
				let  varmap f1 f2 = lterm_ctxmap f1 f2 0
				let  bvar   i = LBCtx i
				let  fvar   i = LFCtx i
			      end)

module CtxbindModal = Binding(struct
				type t = lmodalterm
				type tvar = lctxdesc
				type tenv = int
				let  bound  i = i
				let varmap f1 f2 = modal_ctxmap f1 f2 0
				let bvar i = LBCtx i
				let fvar i = LFCtx i
			      end)

module CtxbindCtxdesc = Binding(struct  type tvar = lctxdesc
				type t = lctxdesc
				type tenv = int
				let  bound  i = i
				let  varmap f1 f2 = tmlist_ctxmap f1 f2 0
				let  bvar   i = LBCtx i
				let  fvar   i = LFCtx i
			end)

module CtxbindLtermS =  (struct
			   let subst_bound_list = CtxbindLterm.subst_bound_list_given_shift CtxbindCtxdesc.shift_bound
			   let subst_bound      = CtxbindLterm.subst_bound_given_list subst_bound_list
			 end)

module CtxbindModalS =  (struct
			   let subst_bound_list = CtxbindModal.subst_bound_list_given_shift CtxbindCtxdesc.shift_bound
			   let subst_bound      = CtxbindModal.subst_bound_given_list subst_bound_list
			 end)

module CtxbindCtxdescS =  (struct
			   let subst_bound_list = CtxbindCtxdesc.subst_bound_list_given_shift CtxbindCtxdesc.shift_bound
			   let subst_bound      = CtxbindCtxdesc.subst_bound_given_list subst_bound_list
			 end)

(*
let inverse_subst meta l =
  let rec is_decreasing_bound curmax subst =
    match subst with
	[] -> true
      | LVar (LBVar hd) :: tl -> hd = curmax - 1 && is_decreasing_bound hd tl
      | _ -> false
  in
  let id = List.map (fun i -> LBMeta(i)) (increasing (List.length l)) in
  let change t (t' : lmodalterm) subst =
    match t with
	LBMeta(i) -> ExtList.updatenth i t' subst
      | LTermInCtx(ctx,LModal(LBMeta(i),s)) when is_decreasing_bound (List.length ctx) s -> ExtList.updatenth i t' subst
      | _ -> subst
  in
  let subst = ExtList.foldindex (fun i (t',t) subst -> change (LBMeta(i)) t' subst) id (List.rev l) in
    match l with [] -> (fun x -> x) | _ -> MetabindLtermS.subst_bound_list (List.rev subst)
*)

let mk_infer ?(guessubst=[]) free bound metafree metabound ctxfree ctxbound =
  let closef free bound cl =
    let envlen = free + bound in
    cl ?howmany:(Some bound) envlen
  in
  let ltop = closef free     bound     BindLterm.close_down in
  let mtop = closef metafree metabound MetabindLterm.close_down in
  let ctop = closef ctxfree  ctxbound  CtxbindLterm.close_down in
  let stop = fun x -> x (* inverse_subst (metafree+metabound) guessubst *) in
  LInfer(mk_lunifvar (), compose stop (compose mtop (compose ctop ltop)))
